DrDamage = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceHook-2.1")
local L = AceLibrary("AceLocale-2.2"):new("DrDamage")
local BS = AceLibrary("Babble-Spell-2.2")
local SEA = AceLibrary("SpecialEvents-Aura-2.0")
local GT = AceLibrary("Gratuity-2.0")

local defaults = {
	--Calculation
	HitCalc = false,
	HitTarget = true,
	TargetLevel = -1,
	TargetPlayer = false,
	SpellDamage = 0,
	Healing = 0,
	ShadowDamage = 0,
	FireDamage = 0,
	NatureDamage = 0,
	FrostDamage = 0,
	ArcaneDamage = 0,
	HolyDamage = 0,
	ItemCrit = 0,
	ItemHit = 0,
	--Actionbar:
	DisplayType = "Avg",
	ManualDmg = false,
	ManualPlus = false,
	FontSize = 8,
	FontPosition = 0,
	CastsLeft = false,
	--Tooltip:
	Display = true,
	PlusDmg = true,
	DispCrit = true,
	DispHit = true,
	AvgHit = true,
	AvgCrit = true,
	Ticks = true,
	Total = true,
	Extra = true,
	Next = true,
	DPS = true,
	DPM = true,
	Doom = true,
}

DrDamage:RegisterDB("DrDamageDB");
DrDamage:RegisterDefaults("profile", defaults )

function DrDamage:OnInitialize()	
end

local changesDMG = {
    "CHARACTER_POINTS_CHANGED",
    "PLAYER_ALIVE",
}

function DrDamage:CHARACTER_POINTS_CHANGED()
    	DrDamage:DmgUpdateTalents()
    	DrDamage:UpdateAB()
end

function DrDamage:PLAYER_ALIVE()
    	DrDamage:DmgUpdateTalents()
    	DrDamage:UpdateAB()
end

local changesAB = {
    "ACTIONBAR_PAGE_CHANGED",
    "ACTIONBAR_SHOWGRID",
    "UNIT_INVENTORY_CHANGED",
    "PLAYER_TARGET_CHANGED",
    "UNIT_AURA",
}

function DrDamage:UNIT_AURA( unit )

	if unit == "target" or unit =="player" then
		
		if not self:IsEventScheduled("UpdatingAB") then
	    		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.5, self)
	    	end
	end
end

function DrDamage:ACTIONBAR_PAGE_CHANGED()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self)
	end
end

function DrDamage:ACTIONBAR_SHOWGRID()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self)
	end
end

function DrDamage:PLAYER_TARGET_CHANGED()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self)
	end
end

function DrDamage:UNIT_INVENTORY_CHANGED( unit )
	
	if unit == "player" then
		if not self:IsEventScheduled("UpdatingAB") then
			self:ScheduleEvent("UpdatingAB", self.UpdateAB, 2, self)
		end
	end
end

function DrDamage:Compatible()

	_, englishClass = UnitClass("player");
		
	if englishClass == "WARLOCK" or englishClass =="MAGE" or englishClass == "PRIEST" or englishClass == "SHAMAN" then
		return true
	else
		return false
	end
end
		  
function DrDamage:OnEnable()

	if not DrDamage:Compatible() then
		--LOCALE:
		self:Print( "Your class is not supported by Dr. Damage." )
		do return end
	end
	
	DrDamage:Hook( GameTooltip, "SetAction", true)
	DrDamage:SecureHook( GameTooltip, "SetSpell" )
	DrDamage:DmgUpdateTalents()
	
    	for i = 1, #changesDMG do
    		
    		if changesDMG[i] then
    			self:RegisterEvent( changesDMG[i] )
    		end
  	end
	
	if DrDamage.db.profile.DisplayType ~= "Off" then
    		
    		DrDamage:UpdateAB()
    	
    		for i = 1, #changesAB do
    	
    			if changesAB[i] then
    				self:RegisterEvent( changesAB[i] )
    			end
    		end
  	end   
end

function DrDamage:OnDisable()

	DrDamage:UpdateAB( true )
	talents = nil
	
end

function dGet (n)
	return function () return DrDamage.db.profile[n] 
	end
end
function dSet (n, v)
	return function (v) DrDamage.db.profile[n] = v; DrDamage:UpdateAB(); 
	end
end

local DisplayTypes = { "Avg", "DPS", "DPSC", "Off" }



--LOCALE:
local options = { 
	type='group',
	args = {
		ActionBar = {
			type = "group", desc = "Options for the actionbar", name="ActionBar",
			args = {
				DisplayType = {
					name = "Display",
					type = "text",
					desc = "Choose what to display on the actionbar",
					validate = DisplayTypes,
					get = dGet("DisplayType"),
					set = dSet("DisplayType", value),
					order = 70,
				},
        			Update = {
            				name = "Update",
            				desc = "Forces update to actionbar",
            				type = "execute",
            				order = 75,
            				func = "UpdateAB",
    				}, 
				FontSize = {
					type = 'range',
					name = "Actionbar font size",
					min = 6,
					max = 14,
					step = 1,
					desc = "Set the font size",
					order = 71,
					get = dGet("FontSize"),
					set = dSet("FontSize", value),
				},
				FontPosition = {
					type = 'range',
					name = "Actionbar font position",
					min = -5,
					max = 5,
					step = 1,
					desc = "Set the font position",
					order = 72,
					get = dGet("FontPosition"),
					set = dSet("FontPosition", value),
				}, 
				CastsLeft = {
					type = 'toggle',
					name = "Healing casts left",
					desc = "Toggles amount of casts left to display on actionbar",
					order = 80,
					get = dGet( "CastsLeft" ),
					set = dSet( "CastsLeft", value ),
				},				
			},
		},
		Tooltip = {
			type = "group", desc = "Options for the tooltips", name="Tooltip",
			args = {
				Display = {
					type = 'toggle',
					name = "Tooltip on/off",
					desc = "Toggles the tooltip info on/off",
					order = 50,
					get = dGet( "Display" ),
					set = dSet( "Display", value ),
				},			
				PlusDmg = {
					type = 'toggle',
					name = "Show +dmg/+healing",
					desc = "Toggles displaying of +dmg/+healing",
					order = 52,
					get = dGet( "PlusDmg" ),
					set = dSet( "PlusDmg", value ),
				},
				DispCrit = {
					type = 'toggle',
					name = "Show crit %",
					desc = "Toggles displaying of crit %",
					order = 54,
					get = dGet( "DispCrit" ),
					set = dSet( "DispCrit", value ),
				},
				DispHit = {
					type = 'toggle',
					name = "Show hit %",
					desc = "Toggles displaying of hit %.",
					order = 56,
					get = dGet( "DispHit" ),
					set = dSet( "DispHit", value ),
				},				
				AvgHit = {
					type = 'toggle',
					name = "Show avg + hit range",
					desc = "Toggles displaying of avg hit",
					order = 58,
					get = dGet( "AvgHit" ),
					set = dSet( "AvgHit", value ),
				},
				AvgCrit = {
					type = 'toggle',
					name = "Show avg + crit range",
					desc = "Toggles displaying of avg crit",
					order = 60,
					get = dGet( "AvgCrit" ),
					set = dSet( "AvgCrit", value ),
				},
				Ticks = {
					type = 'toggle',
					name = "Show per tick/hit",
					desc = "Toggles displaying of per hit/tick values",
					order = 62,
					get = dGet( "Ticks" ),
					set = dSet( "Ticks", value ),
				},
				Total = {
					type = 'toggle',
					name = "Show avg total dmg/heal",
					desc = "Toggles displaying of average in total values",
					order = 64,
					get = dGet( "Total" ),
					set = dSet( "Total", value ),
				},				
				Extra = {
					type = 'toggle',
					name = "Show extra dmg/heal (DoT, Ignite, Chained)",
					desc = "Toggles displaying of extra info",
					order = 66,
					get = dGet( "Extra" ),
					set = dSet( "Extra", value ),
				},
				Next = {
					type = 'toggle',
					name = "Show next values (+1% crit, +10 dmg)",
					desc = "Toggles displaying of +1% crit, +10 dmg values",
					order = 68,
					get = dGet( "Next" ),
					set = dSet( "Next", value ),
				},
				DPS = {
					type = 'toggle',
					name = "Show DPS/HPS",
					desc = "Toggles displaying of DPS/HPS",
					order = 70,
					get = dGet( "DPS" ),
					set = dSet( "DPS", value ),
				},
				DPM = {
					type = 'toggle',
					name = "Show DPM/HPM",
					desc = "Toggles displaying of DPM/HPM",
					order = 72,
					get = dGet( "DPM" ),
					set = dSet( "DPM", value ),
				},
				Doom = {
					type = 'toggle',
					name = "Show Damage/healing until OOM",
					desc = "Toggles displaying of damage/healing until OOM",
					order = 74,
					get = dGet( "Doom" ),
					set = dSet( "Doom", value ),
				},				
			},
		},		
		Calculation = {
			type = "group", desc = "Calculation options", name="Calculation",
			args = {
				HitCalc = {			
					type = 'toggle',
					name = "+Hit calculation",
					desc = "Toggles +hit calculation into averages on/off",
					order = 65,
					get = dGet( "HitCalc" ),
					set = dSet( "HitCalc", value ),
				},
				HitTarget = {			
					type = 'toggle',
					name = "Hit calculation by target level",
					desc = "Toggles +hit calculation by target level.",
					order = 66,
					get = dGet( "HitTarget" ),
					set = dSet( "HitTarget", value ),
				},
				TargetLevel = {
					type = 'range',
					name = "Manual target level setting",
					desc = "Manual set of target level compared to your level. Disable hit by target level.",
					min = -10,
					max = 10,
					step = 1,
					order = 67,
					get = dGet( "TargetLevel" ),
					set = dSet( "TargetLevel", value ),
				},
				TargetPlayer = {
					type = 'toggle',
					name = "Manual target is a player.",
					desc = "Toggles if your manually set target is a player.",
					order = 68,
					get = dGet( "TargetPlayer" ),
					set = dSet( "TargetPlayer", value ),
				},				
				ManualDmg = {
					type = 'toggle',
					name = "Manual variables",
					desc = "Allows you to manually set damage/healing properties",
					order = 71,
					get = dGet( "ManualDmg" ),
					set = dSet( "ManualDmg", value ),
				},
				ManualPlus = {
					type = 'toggle',
					name = "Manual damage adds to detected",
					desc = "Allows you to set if manual values should add to detected values",
					order = 72,
					get = dGet( "ManualPlus" ),
					set = dSet( "ManualPlus", value ),
				}, 
				Healing = {
					type = 'range',
					name = "Healing",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your +healing",
					order = 79,
					get = dGet("Healing"),
					set = dSet("Healing", value),
				},				
				SpellDamage = {
					type = 'range',
					name = "Spell Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your spell damage",
					order = 80,
					get = dGet("SpellDamage"),
					set = dSet("SpellDamage", value),
				},		
				FireDamage = {
					type = 'range',
					name = "Fire Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your fire +damage",
					order = 81,
					get = dGet("FireDamage"),
					set = dSet("FireDamage", value),		
		
				},
				ShadowDamage = {
					type = 'range',
					name = "Shadow Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your shadow +damage",
					order = 82,
					get = dGet("ShadowDamage"),
					set = dSet("ShadowDamage", value),		
				},
				FrostDamage = {
					type = 'range',
					name = "Frost Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your frost +damage",
					order = 83,
					get = dGet("FrostDamage"),
					set = dSet("FrostDamage", value),		
				},
				NatureDamage = {
					type = 'range',
					name = "Nature Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your nature +damage",
					order = 84,
					get = dGet("NatureDamage"),
					set = dSet("NatureDamage", value),		
				},
				ArcaneDamage = {
					type = 'range',
					name = "Arcane Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your arcane +damage",
					order = 85,
					get = dGet("ArcaneDamage"),
					set = dSet("ArcaneDamage", value),		
				},
				HolyDamage = {
					type = 'range',
					name = "Holy Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your holy +damage",
					order = 85,
					get = dGet("HolyDamage"),
					set = dSet("HolyDamage", value),		
				},					
				ItemCrit = {
					type = 'range',
					name = "Item Crit Bonus",
					min = 0,
					max = 1,
					isPercent = true,
					desc = "Input your crit bonus from items",
					order = 88,
					get = dGet("ItemCrit"),
					set = dSet("ItemCrit", value),					
				},
				ItemHit = {
					type = 'range',
					name = "Item Hit Bonus",
					min = 0,
					max = 1,
					isPercent = true,
					desc = "Input your hit bonus from items",
					order = 89,
					get = dGet("ItemHit"),
					set = dSet("ItemHit", value),	
				},
			},
		},
	},
};

DrDamage:RegisterChatCommand( { "/drdmg", "/drdamage" }, options);

local talents = {};

function DrDamage:DmgUpdateTalents()

	for k in pairs( talents ) do 
		talents[k] = nil 
	end 
				
	local talentList = DrDamage:DetermineClass( "Talents" )
	
	if not talentList then
		do return end
	end
	
	local numTabs = GetNumTalentTabs();
	
	for t=1, numTabs do
	
    		local numTalents = GetNumTalents(t);
    		  		
	    	for i=1, numTalents do
	    	
	    		nameTalent, _, _, _, currRank, maxRank= GetTalentInfo(t,i);
	    		
	    	    	if currRank ~= 0 then
	    	    			    	    
	    	    		if talentList[nameTalent] then
	    	    		
	    	    			table.insert( talents, { Name = nameTalent, Rank = currRank } )
	    	    			--self:Print("Talent name: " .. nameTalent .. " Points: " .. currRank .. "/" .. maxRank )
	    	    			
	    	    		elseif GetLocale() ~= "enUS" and GetLocale() ~= "enGB" and BS:HasTranslation( nameTalent ) then
	    	    		
	    	    			for index, talent in pairs( talentList ) do
	    	    			
	    	    				if BS:HasTranslation( index ) then
	    	    				
	    	    					if BS[index] == nameTalent then
	    	    						table.insert( talents, { Name = nameTalent, Rank = currRank } )
	    	    					end
	    	    				end
	    	    			end
	    	    		end    	    
	    	    	end	    	    
	    	end
	end
end


function DrDamage:DetermineClass( requestType )

	_, englishClass = UnitClass("player");
	local ClassTable
	
	if englishClass == "WARLOCK" then
		ClassTable = DrDamageWarlockTable
	elseif englishClass == "MAGE" then
		ClassTable = DrDamageMageTable
	elseif englishClass == "PRIEST" then
		ClassTable = DrDamagePriestTable
	elseif englishClass == "SHAMAN" then
		ClassTable = DrDamageShamanTable
	end

	if ClassTable then
	
		if requestType == "Talents" then
			return ClassTable["talentInfo"]	
		elseif requestType == "Spells" then
			return ClassTable["spellInfo"]
		end
	else
		--LOCALE:
		self:Print( "Unsupported class detected." )	
	end
end

function findpattern(text, pattern, start)
	if text and pattern and string.find( text, pattern, start) then
		return string.sub(text, string.find(text, pattern, start))
	else
		return " "
	end
end

function DrDamage:SetSpell( frame, slot )

	if not DrDamage.db.profile.Display then
		do return end
	end
	
	GT:SetSpell( slot, BOOKTYPE_SPELL )
	local spellName, spellRank = GetSpellName( slot, BOOKTYPE_SPELL );
	
	if spellRank and string.len( spellRank ) >= 6 then
		spellRank = tonumber( findpattern( spellRank, "%d+" ))
	end
	
	if not spellRank then
		spellRank = "None"
	end
	
	DrDamage:DoTooltip( frame, spellName, spellRank, true )

end

function DrDamage:SetAction( frame, slot )

	self.hooks[frame].SetAction(frame, slot)
	
	if not DrDamage.db.profile.Display then
		do return end
	end
	
	local gtype, _, _ = GetActionInfo( slot )
	
	if gtype == "spell" then
	
		local spellName = getglobal("GameTooltipTextLeft1"):GetText()
		local spellRank = getglobal("GameTooltipTextRight1"):GetText()
	
		if spellRank and string.len( spellRank ) >= 6 then
			spellRank = tonumber( findpattern( spellRank, "%d+" ))
		else
			spellRank = "None"
		end
	
		GT:SetAction( slot )
		DrDamage:DoTooltip( frame, spellName, spellRank, true )
		
	elseif gtype == "macro" then
	
  		local spellInfo = DrDamage:DetermineClass( "Spells" )
  		
		if not spellInfo then
			do return end
		end 		

		local macroName = getglobal("GameTooltipTextLeft1"):GetText()
		local spellName
		local spell
				
		if macroName and string.find( macroName, '%(%d+%)') then
			spellName = string.sub( macroName, 1, string.find( macroName, "%(" ) - 1 )
		else
			spellName = macroName
		end
				
		if spellName then
			
			spell = spellInfo[spellName]
			
			if not spell then
					
				for index, cSpell in pairs( spellInfo ) do
						
  					if (type( index ) == "string" ) then
  							
  						if string.find( index, spellName ) then
  								
  							local position = string.find( index, spellName )
  								
  							if position == 1 then
  								spell = cSpell
  								spellName = index
  							end
  						end
					end	
				end
			end
		end
								
		if spell then
										
			local macroRank = findpattern( macroName, '%(%d+%)')
						
			if macroRank then
				macroRank = tonumber( string.sub( macroRank, 2, string.len( macroRank ) - 1 ))
			end
						
			local tableSpell
						
			if macroRank then
				if macroRank > 0 then
					tableSpell = spell[macroRank]
				end
				
				if tableSpell then
					DrDamage:DoTooltip( frame, spellName, macroRank, false )
				end	
			else
				local spellRank = 0
				local playerLevel = UnitLevel("player")
						
				for index, ispell in pairs( spell ) do
						
  					if (type( index ) == "number") and index > spellRank then
  							
  						if ispell.spellLevel <= playerLevel then
							spellRank = index
						end
     					end
				end
							
				if spellRank > 0 then
					tableSpell = spell[spellRank]
				end
				
				if tableSpell then
					DrDamage:DoTooltip( frame, spellName, spellRank, false )
				end					
			end	
		end	
	end
end


local TooltipData = {}

function DrDamage:DoTooltip( frame, spellName, spellRank, readTooltip )

	if spellName and spellRank then
  	
  		local spellInfo = DrDamage:DetermineClass( "Spells" )
  		
		if not spellInfo then
			do return end
		end  		
  	
		if spellInfo[spellName] then
		
			local spell = spellInfo[spellName]
			local baseSpell = spell[0]
			local tableSpell = spell[spellRank]
			
			if not tableSpell then
				tableSpell = spell["None"]
			end
			
			if tableSpell and not baseSpell.NoTooltip then
				
				TooltipData = DrDamage:RawNumbers ( tableSpell, spellName )
				local DPM
				local DOOM
				local ManaCost
				local healingSpell = TooltipData.HealingSpell
	
				if readTooltip then
					_, _, ManaCost = GT:Find(L["(%d+) Mana"], 2, 2)
				end				
				
				if ManaCost then
				
					ManaCost = tonumber( ManaCost )
					
					if ManaCost > 0 then
						
						DPM = DrDamage:Round( TooltipData.AvgTotal / ManaCost, 1 )
					
						local PlayerMana = UnitManaMax("player")
						if PlayerMana then
							DOOM = DrDamage:Round( DPM * PlayerMana )
						end
					end					
					
				end
				
				--LOCALE:
				frame:AddLine(" ")
				
				if DrDamage.db.profile.PlusDmg then
					if healingSpell then
						frame:AddDoubleLine("+Healing: ", TooltipData.SpellDmg .. " (x" .. TooltipData.SpellDmgM .. ")", 1, 1, 1, 0, 0.7, 0  )
					else
						frame:AddDoubleLine("+Dmg: ", TooltipData.SpellDmg .. " (x" .. TooltipData.SpellDmgM .. ")", 1, 1, 1, 0, 0.7, 0  )
					end
				end
  				
  				if DrDamage.db.profile.DispCrit and TooltipData.CritRate > 0 then
  					frame:AddDoubleLine("Crit:", TooltipData.CritRate .. "%", 1, 1, 1, 0, 0.7, 0 )
  				end
  				
  				if DrDamage.db.profile.DispHit and not baseSpell.Unresistable then
  					frame:AddDoubleLine("Hit:", TooltipData.HitRate .. "%", 1, 1, 1, 0, 0.7, 0 )
  				end  				
  				
  				if DrDamage.db.profile.AvgHit then
  					frame:AddDoubleLine("Avg (x".. TooltipData.DmgM .. "):", TooltipData.AvgDmg .. " (".. TooltipData.MinDmg .."-".. TooltipData.MaxDmg ..")", 1, 1, 1, 0.8, 0.8, 0.9 )
  				end
  				
  				if DrDamage.db.profile.AvgCrit and TooltipData.AvgCrit > 0 and not baseSpell.sHits then
  					frame:AddDoubleLine("Avg Crit:", TooltipData.AvgCrit .. " (".. TooltipData.MinCrit .."-".. TooltipData.MaxCrit ..")", 1, 1, 1, 0.8, 0.8, 0.9 )
  				end
  				
  				if DrDamage.db.profile.Ticks then 
  					if TooltipData.PerHit then
  						frame:AddDoubleLine( "Hits:", TooltipData.PerHit[2] .. "x ~" .. TooltipData.PerHit[1], 1, 1, 1, 0.8, 0.8, 0.9 )
  					elseif TooltipData.PerTick then
  						frame:AddDoubleLine( "Ticks:", TooltipData.PerTick[2] .. "x ~" .. TooltipData.PerTick[1], 1, 1, 1, 0.8, 0.8, 0.9 )
  					end
  				end
  				
  				if DrDamage.db.profile.Total and TooltipData.AvgTotal > TooltipData.AvgDmg then
  					frame:AddDoubleLine("Avg Total:", TooltipData.AvgTotal, 1, 1, 1, 0.8, 0.8, 0.9  )
  				end
  				
  				if DrDamage.db.profile.Extra then 
  				
  					if TooltipData.ChainDmg > 0 then
  						frame:AddDoubleLine("Avg Chain Effect:", TooltipData.ChainDmg, 1, 1, 1, 0.8, 0.8, 0.9  )
  					end
  				
  					if TooltipData.IgniteDmg > 0 then
  						frame:AddDoubleLine("Ignite:", TooltipData.IgniteDmg, 1, 1, 1, 0.8, 0.8, 0.9 )
  					end
  				
   					if TooltipData.DotDmg > 0 then
  						frame:AddDoubleLine("DoT:", TooltipData.DotDmg, 1, 1, 1, 0.8, 0.8, 0.9 )
  					end
  				end
  				
  				if DrDamage.db.profile.Next then

  					if TooltipData.NextCrit > 0 then
  						if healingSpell then
  							frame:AddDoubleLine("+1% Crit:", "+" .. TooltipData.NextCrit .. " (= " .. DrDamage:Round( ( TooltipData.NextCrit * 10 ) / TooltipData.NextTenDmg, 1 ) .. " +heal)", 1, 1, 1, 0.3, 0.6, 0.5 )	
  						else
  							frame:AddDoubleLine("+1% Crit:", "+" .. TooltipData.NextCrit .. " (= " .. DrDamage:Round( ( TooltipData.NextCrit * 10 ) / TooltipData.NextTenDmg, 1 ) .. " +dmg)", 1, 1, 1, 0.3, 0.6, 0.5 )	
  						end
  					end
  					
  					if DrDamage.db.profile.HitCalc and not healingSpell then
  						frame:AddDoubleLine( "+1% Hit:", "+" .. TooltipData.NextHit, 1, 1, 1, 0.3, 0.6, 0.5 )
  					end
  				
  					if healingSpell then
  						frame:AddDoubleLine( "+10 Healing:", "+" .. TooltipData.NextTenDmg, 1, 1, 1, 0.3, 0.6, 0.5 )
  					else
  						frame:AddDoubleLine( "+10 Dmg:", "+" .. TooltipData.NextTenDmg, 1, 1, 1, 0.3, 0.6, 0.5 )
  					end
  					
  				end
  				
  				if DrDamage.db.profile.DPS and not baseSpell.NoDPS then
  				
  					if TooltipData.DPSC ~= TooltipData.DPS and not baseSpell.NoDPS then
  						if healingSpell then
  							frame:AddDoubleLine("HPS/HPSC:", TooltipData.DPS .. "/" .. TooltipData.DPSC, 1, 1, 1, 0.8, 0.1, 0.1)
  						else
  							frame:AddDoubleLine("DPS/DPSC:", TooltipData.DPS .. "/" .. TooltipData.DPSC, 1, 1, 1, 0.8, 0.1, 0.1)
      						end
      					else
      				
      						if healingSpell then
      							frame:AddDoubleLine("HPS:", TooltipData.DPS, 1, 1, 1, 0.8, 0.1, 0.1)
      						
      							if TooltipData.DPSCD > 0 then
      								frame:AddDoubleLine("HPS (CD):", TooltipData.DPSCD, 1, 1, 1, 0.8, 0.1, 0.1)
      							end
      						else
      							frame:AddDoubleLine("DPS:", TooltipData.DPS, 1, 1, 1, 0.8, 0.1, 0.1)
      						
      							if TooltipData.DPSCD > 0 then
      								frame:AddDoubleLine("DPS (CD):", TooltipData.DPSCD, 1, 1, 1, 0.8, 0.1, 0.1)
      							end      						
      						end
      					end
      				end
      				
      				if DrDamage.db.profile.DPM and DPM then
      					if healingSpell then
      						frame:AddDoubleLine("HPM:", DPM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					else
      						frame:AddDoubleLine("DPM:", DPM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					end
      				end
      				
      				if DrDamage.db.profile.Doom and DOOM and not baseSpell.NoDoom then
      					if healingSpell then
      						frame:AddDoubleLine("Healing until OOM:", DOOM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					else
      						frame:AddDoubleLine("Damage until OOM:", DOOM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					end
      				end
      				
				for k in pairs( TooltipData ) do 
					TooltipData[k] = nil 
				end      				
      				
      				frame:Show()			
			end  			
		end	  
	end
end
	

local ABData = {}
local ABTextChildren = {}

function DrDamage:UpdateAB( disable )

	local blizzAction
	local BT3 = false
	local Bongos = false
	local DAB = false
	local NAB = false
	local Trinity = false
	local blizzAB = false

	if IsAddOnLoaded("Bartender3") then
		BT3 = true
	elseif IsAddOnLoaded( "Bongos" ) then
		Bongos = true
	elseif IsAddOnLoaded("DiscordActionBars") then
		DAB = true
	elseif IsAddOnLoaded("Nurfed_ActionBars") then
		NAB = true
	elseif IsAddOnLoaded("TrinityBars") then
		Trinity = true		
	else
		blizzAB = true
		blizzAction = { "ActionButton", "BonusActionButton", "MultiBarBottomLeftButton", "MultiBarBottomRightButton","MultiBarRightButton", "MultiBarLeftButton", }
	end
	
	local spellInfo = DrDamage:DetermineClass( "Spells" )
	
	if not spellInfo then
		do return end
	end
	
	for i = 1, 120 do
	
		local ab
		local name
	
		if BT3 then
			name = "BT3Button"..i
			ab = Bartender3.actionbuttons
		elseif Bongos then
			name = "BActionButton"..i
		elseif DAB then
			name = "DAB_ActionButton_"..i
		elseif NAB then
			name = "Nurfed_ActionButton"..i
		elseif Trinity then
			name = "TrinityActionButton"..i
		elseif blizzAB then
			local abType = 1 + math.floor( ( i - 1 ) / 12 )
			local buttonNr = i - math.floor( ( i - 1 ) / 12 ) * 12
			
			if abType > #blizzAction then
				do return end
			end
			
			name = blizzAction[abType] .. buttonNr
		end
					
		local button = getglobal(name)
					
		if button then
			ABTextChildren = { button:GetChildren() }
		
			if ABTextChildren then
   				for i = 1, #ABTextChildren do
   				
   					local frame = ABTextChildren[i]
   		
   					if frame and frame.text and frame:GetID() == 200 then
   						frame.text:Hide()
   					end
   					
   					ABTextChildren[i] = nil
   					
   				end
   			end			
		end
			
		if button and button:IsVisible() and not disable then
		
			local id
		
			if BT3 and ab[i] and ab[i].state == "used" then
				id = ab[i].object:PagedID()
			elseif Bongos then
				id = button:GetPagedID()
			elseif DAB then
				id = button:GetActionID()
			elseif NAB or Trinity then
				id = button:GetID()
			elseif blizzAB then
				id = ActionButton_GetPagedID( button )
			end
			
			if id then
			
				local gtype, pid, _ = GetActionInfo(id)
			
				if gtype == "spell" then
			
					local spellName, spellRank = GetSpellName( pid, BOOKTYPE_SPELL );
			
					if spellInfo[spellName] then
					
						local spell = spellInfo[spellName]
						local tableSpell = spell[ tonumber( findpattern( spellRank, "%d+" ) ) ]
						
						if not tableSpell then
							tableSpell = spell["None"]
						end
		
						if tableSpell then

							ABData = DrDamage:RawNumbers ( tableSpell, spellName )
							
							local didCasts = false
							local baseSpell = spell[0]
							local healingSpell = ABData.HealingSpell
							
							if DrDamage.db.profile.CastsLeft and healingSpell then
							
								GT:SetAction( id )
								local _, _, ManaCost = GT:Find(L["(%d+) Mana"], 2, 2)
								local PlayerMana = UnitManaMax("player")
							
								if ManaCost then
									ManaCost = tonumber( ManaCost )
									local castsLeft = DrDamage:Round( PlayerMana / ManaCost )
									DrDamage:CreateABtext( button, castsLeft, healingSpell )
									didCasts = true
								end
								
							end
							
							if not didCasts then
							
								local noDPS = baseSpell.NoDPS
								local displayType = DrDamage.db.profile.DisplayType
							
								if displayType == "DPS" and not noDPS then
									DrDamage:CreateABtext( button, ABData.DPS, healingSpell )
								elseif displayType == "DPSC" and not noDPS then
									DrDamage:CreateABtext( button, ABData.DPSC, healingSpell )
								elseif displayType == "Avg" or noDPS and displayType ~= "Off" then
									DrDamage:CreateABtext( button, ABData.AvgTotal, healingSpell )
								end
							end
							
							for k in pairs( ABData ) do 
								ABData[k] = nil 
							end
						end				
					end
					
				elseif gtype == "macro" then
				
					local macroName = GetActionText( id )
					local spellName
					local spell
				
					if macroName and string.find( macroName, '%(%d+%)') then
						spellName = string.sub( macroName, 1, string.find( macroName, "%(" ) - 1 )
					else
						spellName = macroName
					end
					
					if spellName then
						
						spell = spellInfo[spellName]
					
						if not spell then
					
							for index, cSpell in pairs( spellInfo ) do
						
  								if (type( index ) == "string" ) then
  							
  									if string.find( index, spellName ) then
  								
  										local position = string.find( index, spellName )
  									
  										if position == 1 then
  											spell = cSpell
  											spellName = index
  										end
  									end
								end	
							end
						end
					end
								
					if spell then
					
						local macroRank = findpattern( macroName, '%(%d+%)')
						
						if macroRank then
							macroRank = tonumber( string.sub( macroRank, 2, string.len( macroRank ) - 1 ))
						end
						
						local tableSpell
						
						if macroRank then
							if macroRank > 0 then
								tableSpell = spell[macroRank]
							end
						else
							local spellRank = 0
							local playerLevel = UnitLevel("player")
						
							for index, ispell in pairs( spell ) do
						
  								if (type( index ) == "number") and index > spellRank then
  							
  									if ispell.spellLevel <= playerLevel then
										spellRank = index
									end
     								end
							end
							
							if spellRank > 0 then
								tableSpell = spell[spellRank]
							end
						end
						
						if tableSpell then
						
							local macroText = getglobal(button:GetName() .. "Name")
					
							if macroText then
								macroText:Hide()
							end						
						
							ABData = DrDamage:RawNumbers ( tableSpell, spellName )
							local displayType = DrDamage.db.profile.DisplayType
														
							if displayType == "DPS" then
								DrDamage:CreateABtext( button, ABData.DPS, ABData.HealingSpell )
							elseif displayType == "DPSC" then
								DrDamage:CreateABtext( button, ABData.DPSC, ABData.HealingSpell )
							elseif displayType == "Avg" then
								DrDamage:CreateABtext( button, ABData.AvgTotal, ABData.HealingSpell )
							end
							
							for k in pairs( ABData ) do 
								ABData[k] = nil 
							end
						end				
					end							
				end
			end
		end
	end	
end


function DrDamage:CreateABtext( button, cDmg, HealingSpell )

	ABTextChildren = { button:GetChildren() }
	local found = false
	local positionMod = 0
	
	if cDmg and string.len( cDmg ) then
		positionMod = math.floor( (4 - string.len( cDmg ) ) * 3.5 )
	end
	
	if ABTextChildren then
   		for i = 1, #ABTextChildren do
   		
   			local frame = ABTextChildren[i]
   			
   			if frame and frame.text and frame:GetID() == 200 then
   			
   				frame.text:SetPoint("BOTTOMLEFT", frame, "BOTTOM", (DrDamage.db.profile.FontPosition - 15) + positionMod, 5)
   				frame.text:SetFont( STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE" )
   				frame.text:SetText( cDmg )
   				frame.text:Show()
   				found = true
   			end
   			
   			ABTextChildren[i] = nil
   		end
   	end
   		
   	if not found then
   		
		local dpsText = CreateFrame("Frame", "DrDamageText", button )
		dpsText:SetAllPoints( button )
		dpsText:SetFrameLevel(dpsText:GetFrameLevel() + 1)
		dpsText:SetID( 200 )
		dpsText.text = dpsText:CreateFontString( nil, "OVERLAY" )
		dpsText.text:SetFontObject(GameFontNormalSmall)
		dpsText.text:SetPoint("BOTTOMLEFT", dpsText, "BOTTOM", (DrDamage.db.profile.FontPosition - 15) + positionMod, 5)
		dpsText.text:SetFont(STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE")
		
		if HealingSpell then
			dpsText.text:SetTextColor(0.4, 1.0, 0.3)
		else
			dpsText.text:SetTextColor(1, 1, 0.2)
		end
		
		dpsText.text:SetText( cDmg )
		dpsText.text:Show()
	end

end

function DrDamage:SpellCrit()

	local crit = GetSpellCritChanceFromIntellect( "player" ) + GetCombatRatingBonus(11)
	return crit
end


local hitDataMOB = { 95, 94, 83, 72, 61, 50, 39, 28, 17, 6 }
local hitDataPlayer = { 95, 94, 87, 80, 73, 66, 59, 52, 45, 38 }
local lastTargetLevel
local lastSpellHit

function DrDamage:SpellHit()

	local targetLevel = UnitLevel( "target" )
	local playerLevel = UnitLevel( "player" )
	local hitPerc = 99
	
	if DrDamage.db.profile.HitTarget then
		
		local levelDiff = 0

		if playerLevel >= 60 and targetLevel == -1 then
			levelDiff = 3
		elseif targetLevel == -1 then
			levelDiff = 10
		elseif targetLevel > playerLevel then
			levelDiff = targetLevel - playerLevel
		elseif targetLevel == playerLevel then
			hitPerc = 96
		end
		
		if levelDiff > 0 and levelDiff <= 10 then
		
			if UnitIsPlayer( "target" ) then
				hitPerc = hitDataPlayer[levelDiff]
			else
				hitPerc = hitDataMOB[levelDiff]
			end
		end
		
		lastTargetLevel = targetLevel
		lastSpellHit = hitPerc
		
	elseif not DrDamage.db.profile.HitTarget then
		
		if DrDamage.db.profile.TargetLevel >= 1 then
			
			if DrDamage.db.profile.TargetPlayer then
					hitPerc = hitDataPlayer[ math.floor( DrDamage.db.profile.TargetLevel ) ]
			else
					hitPerc = hitDataMOB[ math.floor( DrDamage.db.profile.TargetLevel ) ]
			end
					
		elseif math.floor( DrDamage.db.profile.TargetLevel ) == 0 then
			hitPerc = 96
		end
		
		lastTargetLevel = DrDamage.db.profile.TargetLevel
		lastSpellHit = hitPerc		
	end
		
	return hitPerc
end

function DrDamage:GetPlusDamage( school )

	local spellDMG = GetSpellBonusDamage(1)
	
	if school == "Holy" then
  		return GetSpellBonusDamage(2)
	elseif school == "Fire" then
  		return GetSpellBonusDamage(3)
  	elseif school == "Nature" then
  		return GetSpellBonusDamage(4)
  	elseif school == "Frost" then
  		return GetSpellBonusDamage(5)
  	elseif school == "Shadow" then
  		return GetSpellBonusDamage(6)  		
   	elseif school == "Arcane" then
  		return GetSpellBonusDamage(7)
  	else
  		return spellDMG
  	end	
end


function DrDamage:RawNumbers( spell, spellName )

	if not spell or not spellName then
		do return end
	end
		
	local spellInfo = DrDamage:DetermineClass( "Spells" )
	
	if not spellInfo then
		do return end
	end	
	
	local baseSpell = (spellInfo[spellName])[0]
	
	if not talents or #talents < 1 then
		DrDamage:DmgUpdateTalents()
	end
	
	local healingSpell = false
	local spellSchool = baseSpell.School
	local critPerc = DrDamage:SpellCrit() 		--Calculate crit
	local hitPerc = lastSpellHit
	
	--Calculate base hit chance only IF needed:
	local targetLevel = UnitLevel( "target" )
	
	if DrDamage.db.profile.HitTarget and targetLevel ~= lastTargetLevel then
		hitPerc = DrDamage:SpellHit() 		
	elseif not DrDamage.db.profile.HitTarget and DrDamage.db.profile.TargetLevel ~= lastTargetLevel then
		hitPerc = DrDamage:SpellHit()
	end
		
	hitPerc = hitPerc + GetCombatRatingBonus( 8 )   --Add Gear
	local spellDmg
	
	if baseSpell.Healing then
		healingSpell = true
	end
	
	if healingSpell then
		spellDmg = GetSpellBonusHealing() 			-- +healing
	else
		spellDmg = DrDamage:GetPlusDamage( spellSchool ) 	--+all spelldamage
	end
	
			
	if DrDamage.db.profile.HitCalc then
	
	end
	
	
	if DrDamage.db.profile.ManualDmg then
		
		if not DrDamage.db.profile.ManualPlus then
			
			if healingSpell then
				spellDmg = DrDamage.db.profile.Healing
			else
				spellDmg = DrDamage.db.profile.SpellDamage
			end
			
			critPerc = critPerc - GetCombatRatingBonus(11) + DrDamage.db.profile.ItemCrit * 100
			hitPerc = hitPerc - GetCombatRatingBonus(8) + DrDamage.db.profile.ItemHit * 100
		else
			if healingSpell then
				spellDmg = spellDmg + DrDamage.db.profile.Healing
			else
				spellDmg = spellDmg + DrDamage.db.profile.SpellDamage
			end
			
			critPerc = critPerc + DrDamage.db.profile.ItemCrit
			hitPerc = hitPerc + DrDamage.db.profile.ItemHit
		end
		
		if not healingSpell then
		
			if ( spellSchool == "Shadow" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ShadowDamage --add shadowdamage
			elseif ( spellSchool == "Fire" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FireDamage --add firedamage
			elseif ( spellSchool == "Nature" ) then
				spellDmg = spellDmg + DrDamage.db.profile.NatureDamage --add naturedamage
			elseif ( spellSchool == "Frost" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FrostDamage --add frostdamage
			elseif ( spellSchool == "Arcane" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ArcaneDamage --add arcanedamage
			elseif ( spellSchool == "Holy" ) then
				spellDmg = spellDmg + DrDamage.db.profile.HolyDamage --add holydamage			
			end
		else
			spellDmg = spellDmg + DrDamage.db.profile.Healing
		end
	end
					
	local critM = 0.5			--Crit multiplier
	local bNukeDmg = 0			--Modifier for spells with a DOT portion
	local dmgM = 1				--Damage multiplier (total)
	local bDmgM = 1				--Base damage Modifier
	local spellDmgM				--Spell damage multiplier
	local igniteM = 0			--Ignites % of total
	local cooldown = 0			--Spell's cooldown
	local finalMod = 0			--Modifier to final damage +/-
	local MDRank = 0
	
	if baseSpell.Cooldown then
		cooldown = baseSpell.Cooldown
	end

	local castTime = 1.5
	
	if spell.castTime then
		castTime = spell.castTime
	elseif baseSpell.castTime then
		castTime = baseSpell.castTime
	end
	
	local eDuration = 1
	
	if spell.eDuration then
		eDuration = spell.eDuration
	elseif baseSpell.eDuration then
		eDuration = baseSpell.eDuration
	end
	
	if baseSpell.bonusFactor then
		spellDmgM = baseSpell.bonusFactor
	elseif baseSpell.eDot then 
		spellDmgM = ( eDuration / 15)
	else
		spellDmgM = ( castTime / 3.5) 
	end
		
	local talentEffects = DrDamage:DetermineClass( "Talents" )
	
	if not talentEffects then
		do return end
	end
			
	for i, charTalent in ipairs( talents ) do
		
		local talentName = charTalent.Name
		local talentRank = charTalent.Rank
		
		if talentName then
		
			local translatedTalent
			
			if GetLocale() ~= "enUS" and GetLocale() ~= "enGB" and BS:HasTranslation( talentName ) then
			
				for index, talent in pairs( talentEffects ) do
				
					if BS:HasTranslation( index ) and BS[index] == talentName then
						translatedTalent = talent
					end
				end
			end
			
					
			if talentEffects[talentName] or translatedTalent then
				
				local effectTalent
				
				if translatedTalent then
					effectTalent = translatedTalent
				else
					effectTalent = talentEffects[talentName]
				end
				
				function DoCalc( ModType, talentSpell, talentValue, multiply )
				
					if talentSpell == "All" or talentSpell == spellName or BS:HasTranslation( talentSpell ) and BS[talentSpell] == spellName then
					
						if ModType == "SpellDamage" then
						
							if multiply then
								spellDmgM = spellDmgM * ( 1 + talentValue )
							else
								spellDmgM = spellDmgM + talentValue
							end
						elseif ModType == "CastTime" then
							
							if baseSpell.CastMod then
								castTime = castTime - baseSpell.CastMod * talentValue
							else
								castTime = castTime - talentValue
							end
								
							if castTime < 1.5 then
								castTime = 1.5
							end
						elseif ModType == "BaseNuke" then
							bNukeDmg = bNukeDmg + talentValue
						elseif ModType == "BaseDamage" then
							bDmgM = bDmgM + talentValue
						elseif ModType == "Crit" then
							critPerc = critPerc + talentValue
						elseif ModType == "Hit" then
							hitPerc = hitPerc + talentValue
						elseif ModType == "CritMultiplier" then
							critM = critM + talentValue
						elseif ModType == "Ignite" then
							igniteM = talentValue
						elseif ModType == "SpellDuration" then
							eDuration = eDuration + talentValue
						elseif ModType == "Cooldown" then
							cooldown = cooldown - talentValue
						elseif ModType == "MD" then
							MDRank = talentValue
						elseif not ModType then
						
							if multiply then
								dmgM = dmgM * ( 1 + talentValue )
							else
								dmgM = dmgM + talentValue	
							end
						end
					end				
				
				end
						
				local modType = effectTalent.ModType
				local multiply = effectTalent.Multiply
				local talentValue = talentRank * effectTalent.Effect
				
				for j, talentSpell in ipairs( effectTalent.Spells ) do
					DoCalc( modType, talentSpell, talentValue, multiply  )
				end
									
				if effectTalent.Spells2 then
				
					modType = effectTalent.ModType2
					multiply = effectTalent.Multiply2
					
					if effectTalent.Effect2 then
						talentValue = talentRank * effectTalent.Effect2
					end
					
					for j, talentSpell in ipairs( effectTalent.Spells2 ) do
						DoCalc( modType, talentSpell, talentValue, multiply )
					end
				end	
			end
		end		
	end
	
	--Start calculating Buffs/Debuffs that affects damage not provided by blizzard API (What is missing?)
	
	if not healingSpell and DamageBuffs and DamageDebuffs then
	
		for buffName, index in SEA:DebuffIter("target") do
	
			if DamageDebuffs[buffName] and not baseSpell.NoDebuffs then
			
				local buffData = DamageDebuffs[buffName]
			
				if buffData.School and buffData.School == spellSchool or buffData.School2 and buffData.School2 == spellSchool or not buffData.School then
				
					local modType = buffData.ModType
					
					if buffData.Ranks then
						local _, _, _, rank = SEA:UnitHasDebuff( "target", buffName )
						local debuffRank
				
						if rank and string.len( rank ) >= 6 then
							debuffRank = tonumber( findpattern( rank, "%d+" ) )
						end
							
						if not debuffRank then
							debuffRank = buffData.Ranks
						end
						
						if not modType then
							dmgM = dmgM + debuffRank * buffData.Value
						end
					elseif buffData.Apps then
						local _, apps, _, _ = SEA:UnitHasDebuff( "target", buffName )
						
						if modType == "Crit" and apps then
							critPerc = critPerc + apps * buffData.Value
						elseif not modType and apps then
							dmgM = dmgM + apps * buffData.Value
						end
					elseif buffData.Apps2 then
						local _, apps, _, _ = SEA:UnitHasDebuff( "target", buffName )
						dmgM = dmgM + apps * buffData.Value2
					elseif not modType then
						dmgM = dmgM + buffData.Value
					end	
					
					if buffData.Texture and modType == "Texture" then
						local _, _, texture, _ = SEA:UnitHasDebuff( "target", buffName )
						if string.find( texture, buffData.Texture ) then
							dmgM = dmgM + buffData.Value
						end
					end

				end
			end
		end
		
		for buffName, index in SEA:BuffIter("player") do
		
			if DamageBuffs[buffName] and not baseSpell.NoBuffs then
			
				local buffData = DamageBuffs[buffName]
			
				if buffData.School and buffData.School == spellSchool or buffData.School2 and buffData.School2 == spellSchool or not buffData.School then
				
					local modType = buffData.ModType
					
					if buffData.Ranks then
						local _, _, _, rank = SEA:UnitHasBuff( "player", buffName )
						local buffRank
				
						if rank and string.len( rank ) >= 6 then
							buffRank = tonumber( findpattern( rank, "%d+" ) )
						end				
			
						if not buffRank then
							buffRank = buffData.Ranks
						end
			
						if not modType then
							dmgM = dmgM + buffRank * buffData.Value
						end
					elseif buffData.Apps then
						local _, apps, _, _ = SEA:UnitHasBuff( "player", buffName )
						
						if not modType and apps then
							dmgM = dmgM + apps * buffData.Value
						end
					elseif modType == "BaseDamage" then
						if buffData.Spell == spellName then
							bDmgM = bDmgM + buffData.Value
						end
					elseif modType == "SpellDamage" then
						spellDmgM = spellDmgM + buffData.Value
					elseif modType == "Crit" then
						critPerc = critPerc + buffData.Value
						
						if critPerc > 100 then
							critPerc = 100
						end
					elseif modType == "Special" then
						
						if buffName == "Burning Adrenaline" then
							castTime = 1.5
							dmgM = dmgM + buffData.Value
							
						elseif buffName == "Master Demonologist" then
							
							if UnitCreatureFamily( "pet" ) then
								local petFamily = UnitCreatureFamily("pet")
								
								if petFamily == "Succubus" and MDRank > 0 then
									dmgM = dmgM + 0.02 * MDRank					
								elseif petFamily == "Felguard" then
									dmgM = dmgM + 0.01 * MDRank
								end
							end
						end						
					elseif not modType then
						dmgM = dmgM + buffData.Value
					end
				end
			end		
		end
		
	elseif healingSpell and HealingTargetBuffs and HealingTargetDebuffs then
	
		for buffName, index in SEA:BuffIter("target") do
	
			if HealingTargetBuffs[buffName] then
			
				local buffData = HealingTargetBuffs[buffName]
				
				if buffData.Spell == spellName or buffData.Spell == "All" then
				
					if buffData.Apps then
						local _, apps, _, _ = SEA:UnitHasBuff( "target", buffName )
						
						if not buffData.ModType and apps then
							dmgM = dmgM * ( 1 + apps * buffData.Value )
						end
					elseif buffData.Ranks then
						local _, _, _, rank = SEA:UnitHasBuff( "target", buffName )
						local buffRank
				
						if rank and string.len( rank ) >= 6 then
							buffRank = tonumber( findpattern( rank, "%d+" ) )
						end
						
						if not buffRank then
							buffRank = buffData.Ranks
						end
						
						if buffData.ModType == "Amplify Magic" then
							spellDmg = spellDmg + buffData.Value[buffRank]
						elseif not buffData.ModType then
							dmgM = dmgM * ( 1 +  buffRank * buffData.Value )
						end					
					elseif not buffData.ModType then
						dmgM = dmgM * ( 1 + buffData.Value )
					end					
				end
			end
		end
		
		for buffName, index in SEA:DebuffIter("target") do		
			
			if HealingTargetDebuffs[buffName] and not baseSpell.NoDebuffs then
			
				local buffData = HealingTargetDebuffs[buffName]
				
				if buffData.Spell == spellName or buffData.Spell == "All" then
				
					if buffData.Apps then
						local _, apps, _, _ = SEA:UnitHasDebuff( "target", buffName )
					
						if buffData.ModType == "Wound Poison" then
							local _, _, _, rank = SEA:UnitHasDebuff( "target", buffName )
							local buffRank
				
							if rank and string.len( rank ) >= 6 then
								buffRank = tonumber( findpattern( rank, "%d+" ) )
							end
							
							if not buffRank then
								buffRank = buffData.Ranks
							end
							
							if apps then
								finalMod = -1 * apps * buffData.Value[buffRank]
							end
							
						elseif not buffData.ModType and apps then
							dmgM = dmgM * ( 1 - apps * buffData.Value )
						end
					elseif not buffData.ModType then
						dmgM = dmgM * ( 1 - buffData.Value )
					end
				end
			end
		end
	end
	
	--Modified factors?
	
	--Split between bonus to dot and nuke
	if baseSpell.hybridFactor then
		spellDmgM = spellDmgM * baseSpell.hybridFactor
	end
	
	--Factor that applies to all ranks of the spell
	if baseSpell.sFactor then
		spellDmgM = spellDmgM * baseSpell.sFactor
	end
	
	--Factor that applies to current rank of the spell
	if spell.sFactor then
		spellDmgM = spellDmgM * spell.sFactor
	end
	
	
	--Factor for downranking
	local lowRankFactor = 1
	
	if spell.maxLevel and spell.maxLevel < 64 then
		
		local playerLevel = UnitLevel("player")
	
			if healingSpell then
				lowRankFactor = ( spell.maxLevel + 6 ) / playerLevel
			else
				lowRankFactor = ( spell.maxLevel + 4 ) / playerLevel
			end
		
			if lowRankFactor < 1 then
				spellDmgM = spellDmgM * lowRankFactor
			else
				lowRankFactor = 1
			end
	end
	
	--Factor for spells under level 20
	local lowLevelFactor = 1
	
	if spell.spellLevel and spell.spellLevel < 20 then
	
		lowLevelFactor = 1 - ((20 - spell.spellLevel) * 0.0375)
		
		if lowLevelFactor < 1 and lowLevelFactor > 0 then
			spellDmgM = spellDmgM * lowLevelFactor
		else
			lowLevelFactor = 1
		end
	end
	
	--Adding to spells base damage after levelups:
	
	local minDam = spell[1]
	local maxDam = spell[2]
	
	if baseSpell.coeffA and baseSpell.coeffB and spell.spellLevel and spell.maxLevel then
	
		local playerLevel = UnitLevel("player")
	
		if playerLevel > spell.spellLevel then
		
			local diff = baseSpell.diff - 1
			local partial = false	
						
			if ( (spell.maxLevel - spell.spellLevel) / baseSpell.diff ) <= 0.6 and spell.spellLevel <= 20 then
				partial = true
			end
		
			if playerLevel >= spell.maxLevel then
			
				if partial then			
					minDam = minDam * ( 1 +  ( spell.maxLevel - spell.spellLevel ) * ((( spell.maxLevel + baseSpell.coeffA ) / ( spell.maxLevel + baseSpell.coeffB ) - 1 ) / diff ))
					maxDam = maxDam * ( 1 +  ( spell.maxLevel - spell.spellLevel ) * ((( spell.maxLevel + baseSpell.coeffA ) / ( spell.maxLevel + baseSpell.coeffB ) - 1 ) / diff ))				
				else
					minDam = minDam * (( spell.maxLevel + baseSpell.coeffA ) / ( spell.maxLevel + baseSpell.coeffB ))
					maxDam = maxDam * (( spell.maxLevel + baseSpell.coeffA ) / ( spell.maxLevel + baseSpell.coeffB ))
				end
				
			else
			
				if ( spell.maxLevel - spell.spellLevel ) ~= diff then
					diff = spell.maxLevel - spell.spellLevel
				end
				
				minDam = minDam * ( 1 +  ( playerLevel - spell.spellLevel ) * ((( spell.maxLevel + baseSpell.coeffA ) / ( spell.maxLevel + baseSpell.coeffB ) - 1 ) / ( diff - 1) ))
				maxDam = maxDam * ( 1 +  ( playerLevel - spell.spellLevel ) * ((( spell.maxLevel + baseSpell.coeffA ) / ( spell.maxLevel + baseSpell.coeffB ) - 1 ) / ( diff - 1) ))
			end
		end
		
		minDam = math.floor( minDam )
		maxDam = math.ceil( maxDam )
	end
	
	function DmgCalc()
	
		--Initialize variables that are calculated
		local calcAvgDmgCrit = 0		--Calculated final damage with crit
		local calcAvgDmg = 0			--Calculated final damage without crit
		local calcMinDmg = 0			--Calculated dmg range: MIN
		local calcMaxDmg = 0			--Calculated dmg range: MAX
		local calcMinCrit = 0
		local calcMaxCrit = 0		
		local calcDotDmg = 0			--Calculated DOT portion (if applicable)
		local igniteDotDmg = 0			--Calculated Ignite DOT Portion
		local avgIgnite	= 0			--Calculated Average Ignite
		local dispSpellDmgM = spellDmgM		--Calculated spelldmg M
		local perHit = 0
		local perTick = 0
	
		
		if baseSpell.sHits then
			calcMinDmg = dmgM * ( bDmgM * minDam * baseSpell.sHits + (  spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * maxDam * baseSpell.sHits + (  spellDmg * spellDmgM ) )
		elseif spell.hybridDotDmg then
			calcMinDmg = ( dmgM + bNukeDmg ) * ( bDmgM * minDam + ( spellDmg * spellDmgM ) )
			calcMaxDmg = ( dmgM + bNukeDmg ) * ( bDmgM * maxDam + ( spellDmg * spellDmgM ) )
		else
			calcMinDmg = dmgM * ( bDmgM * minDam + ( spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * maxDam + ( spellDmg * spellDmgM ) )
		end
				
		--EXAMPLE: Wound Poison
		if finalMod ~= 0 then
		
			calcMinDmg = calcMinDmg + finalMod
			calcMaxDmg = calcMaxDmg + finalMod
			
			if calcMinDmg < 0 then
				calcMinDmg = 0
			end
			
			if calcMaxDmg < 0 then
				calcMaxDmg = 0
			end	
		end
		
		calcAvgDmg = (calcMinDmg + calcMaxDmg) / 2
				
		--This is for Imp SW:P
		if baseSpell.eDot and baseSpell.eDuration and eDuration > baseSpell.eDuration then
			local bonusDotDamage = ( calcAvgDmg / baseSpell.eDuration ) * ( eDuration - baseSpell.eDuration )
			calcAvgDmg = calcAvgDmg + bonusDotDamage
		end
		
		
		--Crit calculation:
		local critBonus = 0
		
		if baseSpell.canCrit then
			calcMinCrit = calcMinDmg + calcMinDmg * critM
			calcMaxCrit = calcMaxDmg + calcMaxDmg * critM
			critBonus = ( critPerc / 100 ) * calcAvgDmg * critM
			calcAvgDmgCrit = calcAvgDmg + critBonus
		else
			calcAvgDmgCrit = calcAvgDmg
			critPerc = 0;
		end

		--Hit calculation:
		if hitPerc > 99 then
			hitPerc = 99
		end
		
		if DrDamage.db.profile.HitCalc and not healingSpell and not baseSpell.Unresistable then		
			calcAvgDmgCrit = ( calcAvgDmgCrit - critBonus ) * ( hitPerc / 100 ) + critBonus
		end		
					
		if spell.hybridDotDmg then
	
			local hybridFactor = 0
			local sFactor = 1
			
				if baseSpell.hybridFactor then
					hybridFactor = baseSpell.hybridFactor
				end
				
				if baseSpell.sFactor then
					sFactor = baseSpell.sFactor * sFactor
				end
				
				if spell.sFactor then
					sFactor = spell.sFactor * sFactor
				end
				
				if baseSpell.hybridDotFactor then
					sFactor = baseSpell.hybridDotFactor * sFactor
				end
		
			calcDotDmg = dmgM * ( spell.hybridDotDmg + ( eDuration / 15 ) * ( 1 - hybridFactor ) * sFactor * lowRankFactor * lowLevelFactor * spellDmg )
			dispSpellDmgM = dispSpellDmgM +  ( eDuration / 15 ) * ( 1 - hybridFactor ) * sFactor * lowRankFactor * lowLevelFactor
		end
		
		if baseSpell.Ignite then
			igniteDotDmg = igniteM * calcAvgDmgCrit
			avgIgnite = ( critPerc / 100 ) * igniteDotDmg
		end
		
		if spell.extraDotDmg then
		
			local xFactor = 0
			
			if baseSpell.extraDotF then
				xFactor = baseSpell.extraDotF
			end
			
			calcDotDmg = calcDotDmg + spell.extraDotDmg + xFactor * lowRankFactor * lowLevelFactor * spellDmg --TODO: Verify factor
			dispSpellDmgM = dispSpellDmgM + xFactor * lowRankFactor * lowLevelFactor
		end
		
		local chainEffect = 0
		
		if baseSpell.chainFactor then
			chainEffect = calcAvgDmgCrit * ( baseSpell.chainFactor + math.pow( baseSpell.chainFactor, 2 ) + math.pow( baseSpell.chainFactor, 3 ) )
		end
					
		--Begin DPS calculations
		
		local calcDPS = 0
		local calcDPT = 0
		local calcDPSCooldown = 0
	
		if baseSpell.eDot then
			calcDPS = calcAvgDmgCrit / eDuration
		else
			if baseSpell.Ignite then
			
				calcDPS = calcAvgDmgCrit / castTime + avgIgnite / 4
				
				if spell.hybridDotDmg or spell.extraDotDmg then
					calcDPS = calcDPS + calcDotDmg / eDuration
				end
				
				calcAvgDmgCrit = calcAvgDmgCrit + avgIgnite + calcDotDmg
				
			elseif spell.hybridDotDmg or spell.extraDotDmg then
				calcDPS = calcAvgDmgCrit / castTime + calcDotDmg / eDuration
				calcAvgDmgCrit = calcAvgDmgCrit + calcDotDmg
			else
				calcDPS = calcAvgDmgCrit / castTime
			end
		end
		
		calcDPT = calcAvgDmgCrit / castTime
		
		if baseSpell.Cooldown then
			
			if baseSpell.eDot then
				calcDPSCooldown = calcAvgDmgCrit / cooldown
			else
				calcDPSCooldown = calcAvgDmgCrit / cooldown
			end
		end
		
		local results = {
		
				MinDmg = 	DrDamage:Round( calcMinDmg ), 
				MaxDmg = 	DrDamage:Round( calcMaxDmg ),
				AvgDmg = 	DrDamage:Round( calcAvgDmg ),
				MinCrit =	DrDamage:Round( calcMinCrit ),
				MaxCrit = 	DrDamage:Round( calcMaxCrit ),
				AvgCrit = 	DrDamage:Round( ( calcMinCrit + calcMaxCrit ) / 2 ),
				AvgTotal = 	calcAvgDmgCrit, 
				CritRate = 	DrDamage:Round( critPerc, 2 ),
				HitRate = 	DrDamage:Round( hitPerc, 2 ),
				DotDmg = 	DrDamage:Round( calcDotDmg ), 
				DPS = 		DrDamage:Round( calcDPS, 1 ), 
				DPSC = 		DrDamage:Round( calcDPT, 1 ), 
				SpellDmg = 	DrDamage:Round( spellDmg ), 
				SpellDmgM = 	DrDamage:Round( dispSpellDmgM, 3 ),
				DmgM = 		DrDamage:Round( dmgM, 3 ),
				IgniteDmg = 	DrDamage:Round( igniteDotDmg ), 
				DPSCD = 	DrDamage:Round( calcDPSCooldown, 1 ), 
				ChainDmg = 	DrDamage:Round( chainEffect ),
		}
		
		local hits
		local ticks
		
		if spell.sHits then
			perHit = calcAvgDmg / spell.sHits
			hits = spell.sHits
		elseif baseSpell.sHits then
			perHit = calcAvgDmg / baseSpell.sHits
			hits = baseSpell.sHits
		end
		
		if spell.sTicks then
			perTick = calcAvgDmg / spell.sTicks
			ticks = spell.sTicks
		elseif baseSpell.sTicks then
			perTick = calcAvgDmg / baseSpell.sTicks
			ticks = baseSpell.sTicks
		end		
		
		if perHit > 0 then
			results.PerHit = { DrDamage:Round( perHit, 1 ), hits }
		end
		
		if perTick > 0 then
			results.PerTick = { DrDamage:Round( perTick, 1 ), ticks }
		end
			
		return results
	end
	
	local results = DmgCalc()
	
	local returnNextCrit = 0
	local returnNextTenDmg = 0
	local returnNextHit = 0
	
	if DrDamage.db.profile.Next then
	
		if baseSpell.canCrit then
			critPerc = critPerc + 1
			local nextCrit = DmgCalc()
			returnNextCrit = DrDamage:Round( nextCrit.AvgTotal - results.AvgTotal, 1 )
			critPerc = critPerc - 1	
		end
	
		spellDmg = spellDmg + 10
		local nextTen = DmgCalc()
		returnNextTenDmg = DrDamage:Round( nextTen.AvgTotal - results.AvgTotal, 1 )
		spellDmg = spellDmg - 10
	
		if DrDamage.db.profile.HitCalc and not healingSpell then
			hitPerc = hitPerc + 1
			local nextHit = DmgCalc()
			returnNextHit = DrDamage:Round( nextHit.AvgTotal - results.AvgTotal, 1 )
			hitPerc = hitPerc - 1
		end
	end
	
	results.AvgTotal = DrDamage:Round( results.AvgTotal )
	results.NextCrit = returnNextCrit
	results.NextTenDmg = returnNextTenDmg
	results.NextHit = returnNextHit
	
	if healingSpell then
		results.HealingSpell = true
	end
	
	return results
end

function DrDamage:Round(x, y)

	local s
	
	if y then 
		x = x * math.pow(10, y) 
	end
	
	if (x > 0) then
		s = math.floor(x+0.5);
  	else
    		s = math.ceil(x-0.5);
  	end
  	
  	if y then 
  		s = s / math.pow(10, y) 
  	end
  	
  	return s
end

